# -*- coding: utf-8 -*-
"""Classes and functions to evaluate embeddings according to the "Similarity" task.

The Similarity task computes the correlation between the similarities of word pairs according to their representation
and according to human-assigned scores.

Datasets available in this module :

* `WS353` for the `WordSim353 <http://alfonseca.org/eng/research/wordsim353.html>`_ dataset
  (Finkelstein et al., 2002) [1]_.
  Also partitioned by [2]_ into :
    * `WS_SIM` : WordSim Similarity
    * `WS_REL` : WordSim Relatedness
* `RG65` for Rubenstein and Goodenough (1965) dataset [3]_
* `RAREWORD` for the Luong et al.'s (2013)
  `Rare Word (RW) Similarity Dataset <https://nlp.stanford.edu/~lmthang/morphoNLM/>`_ [4]_
* `MEN` for the Bruni et al.'s (2012) MEN dataset [5]_
* `MTURK` for the `Radinsky et al.'s (2011) Mechanical Turk dataset` [6]_

References
----------
.. [1] Finkelstein, L., Gabrilovich, E., Matias, Y., Rivlin, E., Solan, Z., Wolfman, G., & Ruppin, E. (2001, April).
       Placing search in context: The concept revisited. In Proceedings of the 10th international conference on World
       Wide Web (pp. 406-414). ACM.
.. [2] Eneko Agirre, Enrique Alfonseca, Keith Hall, Jana Kravalova, Marius Pasca, Aitor Soroa, A Study on Similarity
       and Relatedness Using Distributional and WordNet-based Approaches, In Proceedings of NAACL-HLT 2009.
.. [3] Rubenstein, Herbert, and John B. Goodenough. Contextual correlates of synonymy. Communications of the ACM,
       8(10):627–633, 1965.
.. [4] Luong, T., Socher, R., & Manning, C. D. (2013, August). Better word representations with recursive neural
       networks for morphology. In CoNLL (pp. 104-113).
.. [5] Bruni, E., Boleda, G., Baroni, M., & Tran, N. K. (2012, July). Distributional semantics in technicolor. In
       Proceedings of the 50th Annual Meeting of the Association for Computational Linguistics: Long Papers-Volume 1
       (pp. 136-145). Association for Computational Linguistics.
.. [6] Radinsky, K., Agichtein, E., Gabrilovich, E., & Markovitch, S. (2011, March). A word at a time: computing word
       relatedness using temporal semantic analysis. In Proceedings of the 20th international conference on
       World wide web (pp. 337-346). ACM.


"""
import numpy as np
import scipy.stats
from collections import namedtuple

import mangoes
import mangoes.utils.metrics
from mangoes.evaluation.base import PrintableReport, BaseEvaluation, BaseEvaluator

_Similarity = namedtuple("Similarity", "word_pair gold")


# ####################################################################################################################
# DATASETS

class Dataset(mangoes.evaluation.base.BaseDataset):
    """Class to create a Dataset of word pairs similarities, to be used in Evaluation class

    Examples
    --------
    >>> from mangoes.evaluation.similarity import Dataset
    >>> user_dataset = Dataset("user dataset", ['lion tiger 0.8', 'sun phone 0.1'])

    Predefined datasets are available in this module:

    >>> import mangoes.evaluation.similarity
    >>> ws353 = mangoes.evaluation.similarity.WS353

    """

    @classmethod
    def parse_question(cls, question):
        """
        Examples
        --------
        >>> Dataset.parse_question('lion tiger 0.8')
        Similarity(word_pair=('lion', 'tiger'), gold=0.8)

        Parameters
        ----------
        question: str
            A splittable string with the word pair and a score

        Returns
        -------
        namedtuple

        """
        if isinstance(question, _Similarity):
            return question
        a, b, gold = question.strip().split()
        return _Similarity((a, b), float(gold))


WS353 = Dataset("WS353", "../resources/en/similarity/wordsim353.txt")
WS353_RELATEDNESS = Dataset("WS353 relatedness", "../resources/en/similarity/ws353_relatedness.txt")
WS353_SIMILARITY = Dataset("WS353 similarity", "../resources/en/similarity/ws353_similarity.txt")
MEN = Dataset("MEN", "../resources/en/similarity/men.txt")
MTURK = Dataset("M. Turk", "../resources/en/similarity/mturk.txt")
RAREWORD = Dataset("Rareword", "../resources/en/similarity/rareword.txt")
RG65 = Dataset("RG65", "../resources/en/similarity/rg65.txt")

ALL_DATASETS = (WS353, WS353_RELATEDNESS, WS353_SIMILARITY, MEN, MTURK, RAREWORD, RG65)


# ####################################################################################################################
# EVALUATOR

class Evaluator(BaseEvaluator):
    def __init__(self, representation):
        """Evaluator to predict similarity scores for word pairs according to the given representation

        Parameters
        ----------
        representation: mangoes.Representation
            The Representation to use
        """
        self.representation = representation

    def predict(self, word_pairs, metric=mangoes.utils.metrics.rowwise_cosine_similarity):
        """Predict the similarity scores for the given word pair(s).

        Examples
        --------
        >>> # create a representation
        >>> import numpy as np
        >>> import mangoes
        >>> vocabulary = mangoes.Vocabulary(['lion', 'tiger', 'sun', 'moon', 'phone', 'germany'])
        >>> matrix = np.array([[1, 0], [1, 0.2], [0, 1], [0, 1.2], [0.7, 0.7], [0.7, 0.8]])
        >>> representation = mangoes.Embeddings(vocabulary, matrix)
        >>> # predict
        >>> import mangoes.evaluation.similarity
        >>> evaluator = mangoes.evaluation.similarity.Evaluator(representation)
        >>> evaluator.predict(('lion', 'tiger'))
        array([ 0.98058068])
        >>> evaluator.predict([('lion', 'tiger'), ('sun', 'phone')])
        {('lion', 'tiger'): 0.98058067569092011, ('sun', 'phone'): 0.70710678118654757}

        Parameters
        ----------
        word_pairs: tuple of 2 str or list of tuples of 2 str
            a word pair or a list of word pairs
        metric
            the metric to use to compute the similarity (default : cosine)

        Returns
        -------
        dict
            A dictionary with analogies as keys and the Predictions as values

        """
        if isinstance(word_pairs, tuple):
            w1, w2 = word_pairs
            return metric(self.representation[w1], self.representation[w2])[0]

        words_pairs_indices = np.array([[self.representation.words.index(q[0]),
                                         self.representation.words.index(q[1])]
                                        for q in word_pairs])

        first_terms = self.representation.matrix[words_pairs_indices[:, 0], :]
        second_terms = self.representation.matrix[words_pairs_indices[:, 1], :]

        return {q: p for q, p in zip(word_pairs, metric(first_terms, second_terms))}


# ####################################################################################################################
# EVALUATION
class Evaluation(BaseEvaluation):
    """Class to evaluate a representation on a dataset or a list of datasets

    Both Pearson and Spearman coefficient are given.

    Examples
    --------
    >>> # create a representation
    >>> import numpy as np
    >>> import mangoes
    >>> vocabulary = mangoes.Vocabulary(['lion', 'tiger', 'sun', 'moon', 'phone', 'germany'])
    >>> matrix = np.array([[1, 0], [1, 0.2], [0, 1], [0, 1.2], [0.7, 0.7], [0.7, 0.8]])
    >>> representation = mangoes.Embeddings(vocabulary, matrix)
    >>> # evaluate
    >>> import mangoes.evaluation.similarity
    >>> dataset = Dataset("test", ['lion tiger 0.8', 'sun moon 0.8', 'phone germany 0.3'])
    >>> evaluation = mangoes.evaluation.similarity.Evaluation(representation, dataset)
    >>> evaluation.get_score() # doctest: +NORMALIZE_WHITESPACE
    Score(pearson=Coeff(coeff=-0.40705977800644011, pvalue=0.73310813349301363),
          spearman=Coeff(coeff=0.0, pvalue=1.0), nb=3)
    >>> print(evaluation.get_report()) # doctest: +NORMALIZE_WHITESPACE
                                                                              pearson       spearman
                                                          Nb questions        (p-val)        (p-val)
    ================================================================================================
    test                                                           3/3  -0.407(7e-01)     0.0(1e+00)
    ------------------------------------------------------------------------------------------------


    Parameters
    ----------
    representation: mangoes.Representation
        The representation to evaluate
    datasets: Dataset
        The dataset(s) to use
    lower: bool
        Whether or not the analogies in the dataset should be lowered
    metric
        the metric to use to compute the similarity (default : cosine)
    """
    _Coeff = namedtuple("Coeff", ["coeff", "pvalue"])
    _Score = namedtuple("Score", "pearson spearman nb")

    class _Report(PrintableReport):
        COL = 5
        HEADER = [(("pearson", "spearman"), (3, 3)),
                  (("Nb questions", "(p-val)", "(p-val)"), (3, 3, 3))]
        PREDICTION_HEADER = [(("gold", "score", ""), (3, 3, 6))]

        def _print_score(self, score):
            string = "{:>{width}}".format("{:.3}({:.0e})".format(*score.pearson), width=3 * self.COL)
            string += "{:>{width}}".format("{:.3}({:.0e})".format(*score.spearman), width=3 * self.COL)
            return string

        def _print_prediction(self, similarity, indent):
            line = "{:{width}}".format('    ' * indent + ' '.join(similarity.word_pair),
                                       width=(self.LINE_LENGTH - 12 * self.COL))
            line += "{:>{width}}".format(similarity.gold, width=3 * self.COL)
            line += "{:>{width}}".format("{:.2}".format(self.evaluation.predictions[similarity.word_pair]),
                                         width=3 * self.COL)
            return line + "\n"

    def __init__(self, representation, *datasets, lower=True, metric=mangoes.utils.metrics.rowwise_cosine_similarity):
        super(Evaluation, self).__init__(Evaluator(representation),
                                         *datasets, lower=lower, evaluator_kwargs={'metric': metric})

    def _filter_list_of_questions(self, list_of_questions):
        subset_similarities = []  # word pairs that can be predicted, respecting duplicates and order from original
        unique_word_pairs = set()  # unique word pairs to resolve
        unique_word_pairs_with_gold = set()  # unique word pairs with expected similarity to detect duplicates
        nb_oov = 0  # number of ignored pairs due to OOV terms
        nb_duplicates = 0  # number of duplicates among the filtered pairs

        for similarity in list_of_questions:
            if self.lower:
                similarity = _Similarity((similarity.word_pair[0].lower(), similarity.word_pair[1].lower()),
                                         similarity.gold)

            (a, b), s = similarity

            if a in self.evaluator.representation.words and b in self.evaluator.representation.words:
                subset_similarities.append(similarity)

                if similarity in unique_word_pairs_with_gold:
                    nb_duplicates += 1
                else:
                    unique_word_pairs_with_gold.add(similarity)

                unique_word_pairs.add((a, b))

            else:
                nb_oov += 1

        result = self._FilteredSubset(subset_similarities, len(list_of_questions), nb_oov, nb_duplicates)
        return result, unique_word_pairs, unique_word_pairs_with_gold

    def _score(self, similarities):
        predictions = [self.predictions[similarity.word_pair] for similarity in similarities]
        gold = [similarity.gold for similarity in similarities]
        pearson = self._Coeff(*scipy.stats.pearsonr(predictions, gold))
        spearman = self._Coeff(*scipy.stats.spearmanr(predictions, gold))
        return pearson, spearman



# -*- coding: utf-8 -*-
"""Utility metrics functions.

"""

import logging

import scipy
import numpy as np
import sklearn.metrics

import mangoes.utils.arrays
import mangoes.utils.exceptions

logger = logging.getLogger(__name__)


def rowwise_cosine_similarity(A, B):
    """Compute cosine_similarity between each corresponding rows of A and B

    Parameters
    ----------
    A: matrix-like object
    B: matrix-like object

    Returns
    -------
    list of float
    """
    if len(A.shape) == 1:
        return sklearn.metrics.pairwise.cosine_similarity(A.reshape((1, -1)), B.reshape((1, -1)))

    # TODO : unify sparse and non sparse
    import scipy.sparse
    if scipy.sparse.issparse(A):
        return np.array(
            [sklearn.metrics.pairwise.cosine_similarity(a.toarray().reshape((1, -1)),
                                                        b.toarray().reshape((1, -1)))[0][0]
             for (a, b) in zip(A, B)])

    return np.array(
        [sklearn.metrics.pairwise.cosine_similarity(a.reshape((1, -1)), b.reshape((1, -1)))[0][0]
         for (a, b) in zip(A, B)])


def pairwise_non_negative_cosine_similarity(first, second, normalize=True):
    """Compute non negative cosine similary between all pairs of vectors in matrices

    Parameters
    ----------
    first: matrix-like object
        :class:`mangoes.utils.arrays.Matrix` with n vectors
    second: matrix-like object
        :class:`mangoes.utils.arrays.Matrix` with k vectors
    normalize: bool
        the matrices have to be normalized : if they both are, set this parameter to False

    Returns
    -------
    matrix-like object
        :class:`mangoes.utils.arrays.Matrix` of shape (n x k)
    """
    if normalize:
        first, second = first.normalize(), second.normalize()

    similarities = first.dot(second.T)  # TODO : check if already normalized
    try:
        return mangoes.utils.arrays.Matrix.factory((similarities + 1) / 2)
    except NotImplementedError:
        # adding a nonzero scalar to a sparse matrix is not yet supported
        return mangoes.utils.arrays.Matrix.factory((similarities.todense() + 1) / 2)


def pairwise_cosine_similarity(first, second, normalize=True):
    """Compute cosine similary between all pairs of vectors in matrices

    Parameters
    ----------
    first: matrix-like object
        a mangoes.utils.arrays.Matrix with n vectors
    second: matrix-like object
        a mangoes.utils.arrays.Matrix with k vectors
    normalize: bool
        the matrices have to be normalized : if they both are, set this parameter to False

    Returns
    -------
    matrix-like object
        mangoes.utils.arrays.Matrix of shape (n x k)
    """
    if normalize:
        first, second = first.normalize(), second.normalize()

    similarities = first.dot(second.T)  # TODO : check if already normalized

    import scipy.sparse  # TODO : try/except
    if scipy.sparse.issparse(similarities):
        similarities = similarities.todense()
    return np.array(similarities)


def _earth_mover_distance(d1, d2, cost):
    """Compute Earth Mover's distance between 2 distributions

    Notes
    -----
    This naive implementation is used as a fallback if pyemd is not installed but can't handle large distributions.

    References
    ----------
    Y. Rubner, C. Tomasi, and L. J. Guibas. "A metric for distributions with applications to image databases."
    In IEEE International Conference on Computer Vision, pages 59-66, January 1998.

    Parameters
    ----------
    d1: list
    d2: list
        Distributions to compare, of sizes n1 and n2
    cost: matrix
        Matrix n1 x n2 with the costs to travel from each point in d1 to each point in d2

    Returns
    -------
    (float, matrix)
        The computed distance and the flow matrix

    """

    def _build_eq_constraint_matrix(n1, n2):
        """Build a matrix to compute equality constraints on rows and columns to be used in scipy.optimize.linprog

        Examples
        ---------
        If n1 = n2 = 2
        The (flatten) flow matrix to build is F = [t0 t1
                                                   t2 t3]
        d1 = [x0 x1]
        d2 = [y0 y1]

        The constraints are :
        - on the rows : t0+t1=x0 and t2+t3=x1
        - on the columns : t0+t2=y0 and t1+t3=y1

        So the matrix is : A = [[1 1 0 0],
                                [0 0 1 1],
                                [1 0 1 0],
                                [0 1 0 1]]
        So A.F = [t0 + t1,
                  t2 + t3,
                  t0 + t2,
                  t1 + t3]
        """

        A = np.zeros((n1 + n2, n1 * n2), dtype=float)
        for i in range(n1):
            # the sum of the values of each row of the flow matrix has to be equal to the corresponding value in d1
            for j in range(n1):
                A[i, i * n1 + j] = 1
        for i in range(n2):
            # the sum of the values of each column of the flow matrix has to be equal to the corresponding value in d2
            for j in range(n2):
                A[i + n2, i + n2 * j] = 1
        return A

    n1, n2 = len(d1), len(d2)

    A_eq = _build_eq_constraint_matrix(n1, n2)
    b_eq = np.array([d1, d2], dtype=float).flatten()

    result = scipy.optimize.linprog(cost.flatten(), A_eq=A_eq, b_eq=b_eq, bounds=[(0, None)] * (n1 * n2),
                                    method='simplex')

    if not result.success:
        logger.error("\n" + str(result))
        raise mangoes.utils.exceptions.RuntimeError("Local implementation of earth mover's distance failed. You "
                                                    "should install POT or pyemd.")

    flow_matrix = result.x.reshape((n1, n2))
    return (flow_matrix * cost).sum(), flow_matrix


def word_mover_distance(representation, sentence1, sentence2,
                        stopwords=None, metric="euclidean",
                        return_flow=False, emd=None):
    """Compute the Word Mover's Distance between two phrases

    References
    ----------
    Matt Kusner et al. "From Word Embeddings To Document Distances"

    POT: Python Optimal Transport : http://pot.readthedocs.io
    pyemd : https://github.com/wmayner/pyemd

    Examples
    --------
    >>> representation = mangoes.Embedding(...)
    >>> sentence_obama = 'Obama speaks to the media in Illinois'
    >>> sentence_president = 'The president greets the press in Chicago'
    >>> import nltk.corpus
    >>> stopwords = nltk.corpus.stopwords.words('english')
    >>> distance = mangoes.utils.metrics.word_mover_distance(representation,
    >>>                                                      sentence_obama, sentence_president, stopwords=stopwords)

    Parameters
    ----------
    representation: mangoes.Representation
        Words vectors
    sentence1: str or list of str
    sentence2: str or list of str
        The two sentences, phrases or documents to compare
    stopwords: list of str (optional)
        List of words to ignore
    metric: str
        Metric to use to compute distances between words (see Representation.pairwise_distances())
    return_flow: boolean (optional)
        If True, returns the flow matrix and the corresponding words with the distance
    emd: None (default), "pot", "pyemd" or callable
        Implementation of Earth Mover's distance to use.
        If None, try to import `POT` or `pyemd` and use it. If neither `POT` nor `pyemd` is installed, use the
        implementation in this module.
        You can also use your own implementation (see _earth_mover_distance())

    Returns
    -------
    float or (float, dict)
        Returns the computed word mover distance.
        If return_flow, also returns a dictionary with the outgoing flows from the words in sentence1 to the words
        of sentence2

    """

    def _check_emd(emd):
        if not emd:
            emd = _get_default_emd()

        if callable(emd):
            return emd

        if emd in {"POT", "pot", "ot"}:
            return _emd_with_pot
        elif emd == "pyemd":
            return _emd_with_pyemd
        else:
            return _earth_mover_distance

    def _get_default_emd():
        import importlib
        if importlib.util.find_spec("ot"):
            return 'pot'
        if importlib.util.find_spec("pyemd"):
            return 'pyemd'

        logger.warning(
            "neither pot or pyemd is installed : a local implementation will be used instead but "
            "can't handle large sentences : you should consider installing pot or pyemd")
        return _earth_mover_distance

    def _emd_with_pot(d1, d2, words_distances):
        import ot
        distance, log = ot.emd2(d1, d2, words_distances, return_matrix=True)
        flow_matrix = log['G']
        return distance, flow_matrix

    def _emd_with_pyemd(d1, d2, words_distances):
        import pyemd
        distance, flow_matrix = pyemd.emd_with_flow(d1, d2, words_distances)
        return distance, flow_matrix

    def _prepare_sentence(sentence):
        if isinstance(sentence, str):
            sentence = sentence.split()
        return [w for w in sentence if w in representation.words]

    def _sentence_to_nbow_vector(sentence, words):
        nbow = np.zeros((len(words),))
        for w in sentence:
            try:
                nbow[words.index(w)] += 1
            except KeyError:
                pass  # stopwords
        return nbow / sum(nbow)

    sentence1, sentence2 = _prepare_sentence(sentence1), _prepare_sentence(sentence2)
    words = set(sentence1 + sentence2)
    if stopwords:
        words -= set(stopwords)
    words = mangoes.Vocabulary(list(words))

    if len(sentence1) == 0 or len(sentence2) == 0 or len(words) == 0:
        # TODO : log a warning
        return np.nan

    d1, d2 = _sentence_to_nbow_vector(sentence1, words), _sentence_to_nbow_vector(sentence2, words)

    words_distances = representation.pairwise_distances(words, metric=metric)

    emd = _check_emd(emd)
    distance, flow_matrix = emd(d1, d2, words_distances)

    if not return_flow:
        return distance

    flow = {}
    for w in sentence1:
        if w in words:
            flow[w] = {(words[j], f) for j, f in enumerate(flow_matrix[words.index(w)]) if f}

    return distance, flow



# -*-coding:utf-8 -*
"""Utility functions to read and write data from/to text files

"""
import fileinput
import logging
import os
import os.path
from contextlib import contextmanager

import mangoes.utils
from mangoes.constants import ENCODING

logger = logging.getLogger(__name__)


def recursive_list_files(path):
    """Return list of base_path to the files present inside the directory (and all subdirectories')
    pointed to by 'base_path'.

    Parameters
    ----------
    path: str
        base_path to the folder to scan for files

    Returns
    -------
    list of str
    """
    files = []
    if os.path.isfile(path):
        files.append(path)
    else:
        for entry in sorted(os.listdir(path)):
            if not entry.startswith("."):
                a_file_path = os.path.join(path, entry)
                if os.path.isfile(a_file_path):
                    files.append(a_file_path)
                else:
                    files.extend(recursive_list_files(a_file_path))
    return files


@contextmanager
def get_reader(source):
    """Context manager to get a reader to a source

    Parameters
    ----------
    source : str or Iterable
        If the source is a path to a file, an archive or a directory, yields a reader to this(these) file(s)
        If the source is an iterable, yields it

    Yields
    ------
    Iterable
    """
    def hook_compressed_encoded_text(filename, mode='r', encoding=ENCODING):
        ext = os.path.splitext(filename)[1]
        if ext == '.gz':
            import gzip
            return gzip.open(filename, mode + 't', encoding=encoding)
        else:
            return open(filename, mode, encoding=encoding)

    if isinstance(source, str):
        if not os.path.exists(source):
            raise mangoes.utils.exceptions.ResourceNotFound(path=source)

        try:
            file_path_list = sorted(recursive_list_files(source))
            reader = fileinput.input(file_path_list, mode="r", openhook=hook_compressed_encoded_text)
            yield reader
        finally:
            try:
                reader.close()
            except UnboundLocalError:
                pass  # reader has not been opened yet
    else:
        yield source



# -*- coding: utf-8 -*-
"""
This module provides an interface into the transformers BERT models for pretraining using either masked language
modeling and next sentence prediction, or only masked language modeling.
"""
import warnings

import transformers
import torch

from mangoes.modeling.training_utils import MangoesLineByLineDatasetForNSP, MangoesLineByLineIterableDataset, \
    IterableCompatibleTrainer, MangoesLineByLineDataset
from mangoes.modeling.bert_base import PipelineMixin, TransformerModel


class BERTForPreTraining(PipelineMixin, TransformerModel):
    """Class for pretraining a Bert model (masked language modeling and next sentence prediction)
    This is a base bert architecture with the masked language modeling head and next sentence prediction head.

    :var self.tokenizer: transformers.BertTokenizerFast object,
        see https://huggingface.co/transformers/model_doc/bert.html#berttokenizerfast for more documentation
    :var self.model: transformers.BertForPreTraining object,
        see https://huggingface.co/transformers/model_doc/bert.html#bertforpretraining

    Parameters
    ----------
    pretrained_tokenizer: str
        Either:
            - A string with the `shortcut name` of a pretrained tokenizer to load from cache or download, e.g.,
              ``bert-base-uncased``.
            - A string with the `identifier name` of a pretrained tokenizer that was user-uploaded to our S3, e.g.,
              ``dbmdz/bert-base-german-cased``.
            - A path to a `directory` containing tokenizer vocab or file saved using
              :func:`tokenizer.save_model` or :func:`tokenizer.save`, e.g., ``./my_model_directory/``.
    device: int, optional, defaults to None
        if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
    **keyword_args include arguments passed to transformers.BertConfig (see mangoes.modeling.BERTBase docstring)
    """

    def __init__(self, pretrained_tokenizer, device=None, **keyword_args):
        TransformerModel.__init__(self, device)
        PipelineMixin.__init__(self)
        self.tokenizer = transformers.BertTokenizerFast.from_pretrained(pretrained_tokenizer)
        vocab_size = len(self.tokenizer.get_vocab())
        config = transformers.BertConfig(vocab_size, **keyword_args)
        self.model = transformers.BertForPreTraining(config)

    def _construct_pipeline(self):
        """
        Implementation for creating inference pipeline.
        """
        self.model.eval()
        mlm_model = transformers.BertForMaskedLM(config=self.model.config)
        bert_weights = self.model.bert.state_dict()
        bert_weights.pop('pooler.dense.weight', None)
        bert_weights.pop('pooler.dense.bias', None)
        mlm_model.bert.load_state_dict(bert_weights)
        mlm_weights = self.model.cls.predictions.state_dict()
        mlm_weights_new = {}
        for key in mlm_weights.keys():
            mlm_weights_new["predictions." + key] = mlm_weights[key]
        mlm_model.cls.load_state_dict(mlm_weights_new)
        self.pipeline = transformers.FillMaskPipeline(mlm_model, self.tokenizer,
                                                      device=-1 if self.model.device.type == "cpu" else
                                                      self.model.device.index)

    def predict(self, inputs, text_pairs=None, top_k=5):
        """
        Predicts masked tokens and next sentence prediction scores.

        Parameters
        ----------
        inputs: str or list of strs
            inputs to predict on
        text_pairs: str or list of strs
            The next sentence for each sentence in the text argument, to predict using the next sentence prediction head
            of the model. Optional, if left blank, next sentence prediction outputs don't mean anything.
            If not None, should be same type as inputs, and same length if list.
        top_k: number of predictions to return per masked token

        Returns
        -------
        masked language predictions: list of dict
            for each masked token, a dict containing:
                sequence (str) – The corresponding input with the mask token prediction.
                score (float) – The corresponding probability.
                token_id (int) – The predicted token id (to replace the masked one).
                token_str (str) – The predicted token (to replace the masked one).
        next_sentence_predictions: List of ints, size (batch_size)
            list of next sentence predictions. If input is 1 string, list will have 1 element. 0 indicates sequence B is
            a continuation of sequence A, 1 indicates sequence B is a random sequence.
        """
        if text_pairs:
            assert type(inputs) == type(text_pairs)
            sequences = []
            if isinstance(inputs, list):
                for i in range(len(inputs)):
                    sequences.append(self.tokenizer.cls_token + " " + inputs[i] + " " + self.tokenizer.sep_token + " "
                                     + text_pairs[i] + " " + self.tokenizer.sep_token)
            else:
                sequences.append(self.tokenizer.cls_token + " " + inputs + " " + self.tokenizer.sep_token + " " +
                                 text_pairs + " " + self.tokenizer.sep_token)
            masked_predictions = self._predict(sequences, top_k=top_k, add_special_tokens=False)
        else:
            masked_predictions = self._predict(inputs, top_k=top_k)
        if isinstance(masked_predictions[0], list):
            masked_predictions = [m[0] for m in masked_predictions]
        outputs = self.generate_outputs(inputs, text_pairs=text_pairs)
        return masked_predictions, torch.argmax(outputs["seq_relationship_logits"], dim=1).tolist()

    def generate_outputs(self, text, text_pairs=None, pre_tokenized=False, output_attentions=False,
                         output_hidden_states=False, word_embeddings=False):
        """
        Tokenize input text and pass it through the BERT model, optionally outputting hidden states or attention
        matrices.

        Parameters
        ----------
        text: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]]
            the text to compute features for.
        text_pairs: str or List[str] if pre_tokenized=False, else List[str] or List[List[str]], optional
            The next sentence for each sentence in the text argument, to predict using the next sentence prediction head
            of the model. Optional, if left blank, next sentence prediction outputs don't mean anything.
        pre_tokenized: Boolean
            whether or not the input text is pretokenized (ie, split on spaces)
        output_attentions: Boolean, optional, defaults to False
            Whether or not to return the attentions tensors of all attention layers.
        output_hidden_states: Boolean, optional, defaults to False
            Whether or not to return the hidden states of all layers.
        word_embeddings: Boolean
            Whether or not to filter special token embeddings and average subword embeddings (hidden states) into word
            embeddings. This functionality is not available for this task class. Use the feature extraction class
            instead.

        Returns
        -------
        Dict containing (note that if single text sequence is passed as input, batch size will be 1):
        hidden_states: (Tuple (one for each layer) of torch.FloatTensor (batch_size, sequence_length, hidden_size)).
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
            Only returned if output_hidden_states is True
        attentions: Tuple of torch.FloatTensor (one for each layer) of shape (batch_size, num_heads, sequence_length,
            sequence_length).
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention
            heads.
            Only returned if output_attentions is True
        offset_mappings: Tensor of shape (batch_size, sequence_length, 2)
            Tensor containing (char_start, char_end) for each token, giving index into input strings of start and end
            character for each token. If input is pre-tokenized, start and end index maps into associated word. Note
            that special tokens are included with 0 for start and end indices, as these don't map into input text
            because they are added inside the function. If input includes text pairs, offset mappings for both sentences
            in pair are merged to one row, but indices still align to sentences separately.
        prediction_logits: (torch.FloatTensor of shape (batch_size, sequence_length, config.vocab_size))
            Prediction scores of the language modeling head (scores for each vocabulary token before SoftMax).
        seq_relationship_logits: (torch.FloatTensor of shape (batch_size, 2))
            Prediction scores of the next sequence prediction (classification) head (scores of True/False
            continuation before SoftMax). Only meaningful if text_pairs input argument is non None.
        """
        self.model.eval()
        inputs = self.tokenizer(text, text_pairs, is_split_into_words=pre_tokenized, padding=True, return_tensors='pt',
                                return_offsets_mapping=True)
        offset_mappings = inputs.pop('offset_mapping')
        with torch.no_grad():
            if self.model_device.type == "cuda":
                inputs = {name: tensor.to(self.model_device) for name, tensor in inputs.items()}
                if not next(self.model.parameters()).is_cuda:
                    self.model.to(self.model_device)
            results = self.model.forward(**inputs, return_dict=True, output_attentions=output_attentions,
                                         output_hidden_states=output_hidden_states)
        if word_embeddings:
            warnings.warn("Word embedding consolidation not available for pretraining model. Consider using the"
                          "mangoes.modeling.PretrainedTransformerModelForFeatureExtraction class for feature "
                          "extractions.")
        results["offset_mappings"] = offset_mappings
        return dict(results)

    @classmethod
    def load(cls, pretrained_tokenizer, pretrained_model, device=None):
        """
        load a mangoes BERTForPreTraining object from saved tokenizer and model files.

        Parameters
        ----------
        pretrained_tokenizer: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        pretrained_model: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        device: int, optional, defaults to None
            if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
        """
        model = transformers.BertForPreTraining.from_pretrained(pretrained_model)
        model_object = cls(pretrained_tokenizer, device)
        model_object.model = model
        model_object.model.eval()
        return model_object

    def train(self, output_dir=None, train_text=None, eval_text=None, mlm_probability=0.15,
              short_seq_probability=0.1, nsp_probability=0.5, max_len=None, collator=None, train_dataset=None,
              eval_dataset=None, compute_metrics=None, trainer=None, **training_args):
        """
        Pretrain BERT model on text files

        Input file format:

        (1) One sentence per line. These should ideally be actual sentences, not entire paragraphs or arbitrary spans of
        text.
        (2) Blank lines between documents. Document boundaries are needed so that the "next sentence prediction" task
        doesn't span between documents. Opening a new file will assume a new document as well.
        (3) Assumes different input files contain different documents. (ie, a document cannot span multiple files)


        Example:
        I am very happy. (new line) Here is the second sentence. (new line) (new line) A new document.

        Parameters
        ----------
        output_dir: str
            Path to the output directory where the model predictions and checkpoints will be written. Used to
            instantiate Trainer if trainer argument is None.
        train_text: str or List[str]
            path to training file(s). See above note on formatting for Next Sentence Prediction
        eval_text: (Optional) str or List[str]
            path to evaluation file(s). See above note on formatting for Next Sentence Prediction
        mlm_probability: float
            probability of tokens to be masked while pretraining
        short_seq_probability: float
            probability of short sequences when generating next sentence pairs
        nsp_probability: float
            probability of random next sentence when generating next sentence pairs
        max_len: int
            max length of input sequence. Will default to self.tokenizer.max_length() if None
        collator: Transformers.DataCollator
            custom collator to use
        train_dataset, eval_dataset: torch.Dataset
            instantiated custom dataset object
        compute_metrics: function
            The function that will be used to compute metrics at evaluation. Must return a dictionary string to metric
            values. Used by the trainer, see https://huggingface.co/transformers/training.html#trainer for more info.
        trainer: Transformers.Trainer
            custom instantiated trainer to use
        training_args:
            keyword arguments for training. For complete list, see
            https://huggingface.co/transformers/main_classes/trainer.html#transformers.TrainingArguments
        """
        if not trainer:
            if not output_dir:
                raise RuntimeError("Must provide output directory argument to train() method if trainer argument is "
                                   "None")
            if not collator:
                collator = transformers.DataCollatorForLanguageModeling(self.tokenizer, mlm_probability=mlm_probability)
            if not train_dataset:
                if not train_text:
                    raise RuntimeError("No train text or dataset provided to train method")
                train_dataset = MangoesLineByLineDatasetForNSP(train_text, self.tokenizer,
                                                               short_seq_probability=short_seq_probability,
                                                               nsp_probability=nsp_probability, max_len=max_len)
            if eval_text and not eval_dataset:
                eval_dataset = MangoesLineByLineDatasetForNSP(eval_text, self.tokenizer,
                                                              short_seq_probability=short_seq_probability,
                                                              nsp_probability=nsp_probability, max_len=max_len)
            if eval_dataset is not None and "evaluation_strategy" not in training_args:
                training_args["evaluation_strategy"] = "epoch"
            train_args = transformers.TrainingArguments(output_dir=output_dir, **training_args)
            trainer = transformers.Trainer(self.model, args=train_args, train_dataset=train_dataset,
                                           eval_dataset=eval_dataset, data_collator=collator, tokenizer=self.tokenizer,
                                           compute_metrics=compute_metrics)
        trainer.train()
        self.model.eval()


class BERTForMaskedLanguageModeling(PipelineMixin, TransformerModel):
    """Class for pretraining a Bert model (masked language modeling and next sentence prediction)
    This is a base bert architecture with the masked language modeling head.

    :var self.tokenizer: transformers.BertTokenizerFast object,
        see https://huggingface.co/transformers/model_doc/bert.html#berttokenizerfast for more documentation
    :var self.model: transformers.BertForMaskedLM object,
        see https://huggingface.co/transformers/model_doc/bert.html#bertformaskedlm

    Parameters
    ----------
    pretrained_tokenizer: str
        Either:
            - A string with the `shortcut name` of a pretrained tokenizer to load from cache or download, e.g.,
              ``bert-base-uncased``.
            - A string with the `identifier name` of a pretrained tokenizer that was user-uploaded to our S3, e.g.,
              ``dbmdz/bert-base-german-cased``.
            - A path to a `directory` containing tokenizer vocab or file saved using
              :func:`tokenizer.save_model` or :func:`tokenizer.save`, e.g., ``./my_model_directory/``.
    device: int, optional, defaults to None
        if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
    **keyword_args include arguments passed to transformers.BertConfig (see mangoes.modeling.BERTBase docstring)
    """

    def __init__(self, pretrained_tokenizer, device=None, **keyword_args):
        TransformerModel.__init__(self, device)
        PipelineMixin.__init__(self)
        self.tokenizer = transformers.BertTokenizerFast.from_pretrained(pretrained_tokenizer)
        vocab_size = len(self.tokenizer.get_vocab())
        config = transformers.BertConfig(vocab_size, **keyword_args)
        self.model = transformers.BertForMaskedLM(config)

    def _construct_pipeline(self):
        """
        Implementation for creating inference pipeline.
        """
        self.model.eval()
        self.pipeline = transformers.FillMaskPipeline(self.model, self.tokenizer,
                                                      device=-1 if self.model.device.type == "cpu" else
                                                      self.model.device.index)

    def predict(self, inputs, top_k=5):
        """
        Predicts masked tokens

        Parameters
        ----------
        inputs: str or list of strs
            inputs to fill masked tokens
        top_k: number of predictions to return per masked token

        Returns
        -------
        list of dict, or list of list of dict
        for each masked token, a dict containing:
            sequence (str) – The corresponding input with the mask token prediction.
            score (float) – The corresponding probability.
            token (int) – The predicted token id (to replace the masked one).
            token (str) – The predicted token (to replace the masked one).
        """
        return self._predict(inputs, top_k=top_k)

    @classmethod
    def load(cls, pretrained_tokenizer, pretrained_model, device=None):
        """
        Load a mangoes BERTForMaskedLanguageModeling object from saved tokenizer and model files.

        Parameters
        ----------
        pretrained_tokenizer: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        pretrained_model: str
            Either:
                - A string with the `shortcut name` of a pretrained model to load from cache or download, e.g.,
                  ``bert-base-uncased``.
                - A string with the `identifier name` of a pretrained model that was user-uploaded to our S3, e.g.,
                  ``dbmdz/bert-base-german-cased``.
                - A path to a `directory` containing model weights saved using
                  :func:`~save_pretrained`, e.g., ``./my_model_directory/``.
        device: int, optional, defaults to None
            if -1, use cpu, if >= 0, use CUDA device number. If None, will use GPU if available
        """
        model = transformers.BertForMaskedLM.from_pretrained(pretrained_model)
        model_object = cls(pretrained_tokenizer, device)
        model_object.model = model
        model_object.model.eval()
        return model_object

    def train(self, output_dir=None, train_text=None, eval_text=None, mlm_probability=0.15, iterable_dataset=True,
              max_len=None, collator=None, train_dataset=None, eval_dataset=None, compute_metrics=None, trainer=None,
              **training_args):
        """
        Pretrain BERT model on text files using only masked lm.

        Parameters
        ----------
        output_dir: str, defaults to None
            Path to the output directory where the model predictions and checkpoints will be written. Used to
            instantiate Trainer if trainer argument is None.
        train_text: str or List[str]
            path to training file(s)
        eval_text: (Optional) str or List[str]
            path to evaluation file(s)
        mlm_probability: float
            probability of tokens to be masked while pretraining
        iterable_dataset: Boolean
            when loading dataset during training, whether to process on the fly with no shuffling (True) or load all
            into memory with shuffling (False)
        max_len: int
            max length of input sequence. Will default to self.tokenizer.max_length() if None
        collator: Transformers.DataCollator
            custom collator to use
        train_dataset, eval_dataset: torch.Dataset
            instantiated custom dataset object
        compute_metrics: function
            The function that will be used to compute metrics at evaluation. Must return a dictionary string to metric
            values. Used by the trainer, see https://huggingface.co/transformers/training.html#trainer for more info.
        trainer: Transformers.Trainer
            custom instantiated trainer to use
        training_args:
            keyword arguments for training. For complete list, see
            https://huggingface.co/transformers/main_classes/trainer.html#transformers.TrainingArguments
        """
        if not trainer:
            if not output_dir:
                raise RuntimeError("Must provide output directory argument to train() method if trainer argument is "
                                   "None")
            if not collator:
                collator = transformers.DataCollatorForLanguageModeling(self.tokenizer, mlm_probability=mlm_probability)
            if not train_dataset and iterable_dataset:
                if not train_text:
                    raise RuntimeError("No train text or dataset provided to train method")
                train_dataset = MangoesLineByLineIterableDataset(train_text, self.tokenizer, max_len=max_len)
                if eval_text:
                    eval_dataset = MangoesLineByLineIterableDataset(eval_text, self.tokenizer, max_len=max_len)
                if eval_dataset is not None and "evaluation_strategy" not in training_args:
                    training_args["evaluation_strategy"] = "epoch"
                train_args = transformers.TrainingArguments(output_dir=output_dir, **training_args)
                trainer = IterableCompatibleTrainer(self.model, args=train_args, train_dataset=train_dataset,
                                                    eval_dataset=eval_dataset if eval_text else None,
                                                    data_collator=collator, tokenizer=self.tokenizer)
            else:
                if not train_dataset:
                    if not train_text:
                        raise RuntimeError("No train text or dataset provided to train method")
                    train_dataset = MangoesLineByLineDataset(train_text, self.tokenizer, max_len=max_len)
                if eval_text and not eval_dataset:
                    eval_dataset = MangoesLineByLineDataset(eval_text, self.tokenizer, max_len=max_len)
                if eval_dataset is not None and "evaluation_strategy" not in training_args:
                    training_args["evaluation_strategy"] = "epoch"
                train_args = transformers.TrainingArguments(output_dir=output_dir, **training_args)
                trainer = transformers.Trainer(self.model, args=train_args, train_dataset=train_dataset,
                                               eval_dataset=eval_dataset, data_collator=collator,
                                               tokenizer=self.tokenizer, compute_metrics=compute_metrics)
        trainer.train()
        self.model.eval()



# -*- coding: utf-8 -*-
"""
The modeling subpackage provides functionality for using BERT models for various tasks, including feature extraction,
pretraining, and a number of fine-tuning tasks. It is a interface into the Huggingface transformers and tokenizers
libraries (https://huggingface.co/transformers/ and https://huggingface.co/docs/tokenizers/python/latest/), meant to
provide an easy-to-use API for using BERT for contextual embeddings or training/inference.

The main component of the subpackage are the task classes. Each possible task is implemented as a class that inherits
from :class:`mangoes.modeling.bert_base.TransformerModel`. This API provides full pipeline functionality for each task,
from raw textual inputs to inference, and training, saving, and loading. Each class has a tokenizer and model as
attributes.

The API for these task classes include 5 main functions:
    - load():
        This function will load a pretrained tokenizer and model. Model loading includes loading a pretrained base
        BERT model (while initializing possible fine-tuning head(s)) for use in fine-tuning, or loading a pretrained base
        BERT model and trained classification heads. As such, for all tasks except pretraining from scratch, this is the
        preferred way to instantiate a task class.
        Arguments into this load method can be paths to local directories containing pretrained tokenizers/models, or
        keyword identifiers that map into Huggingface's model hub (see https://huggingface.co/models for more info).

    - save():
        This function is a way to save a trained model and/or tokenizer to a local directory.

    - train():
        The train function provides an easy-to-use interface for training the model for the associated task.
        Under the hood, this function makes use of the transformers.Trainer interface (see
        https://huggingface.co/transformers/training.html#trainer). The train methods in the task classes are meant to have
        intuitive default functionality, while providing users powerful options for customization. With this in mind, there
        are 3 ways to call each task class' train method:

        - raw data:
            Provide raw (ie, not tokenized or featurized) data as input to the method. Dataset classes will be
            instantiated internally using the mangoes default task Dataset classes in
            :module:`mangoes.modeling.training_utils`, which will then be used to instantiated a transformers.Trainer to
            train the model.

        - dataset objects:
            Provide your own dataset objects as input, instead of the raw data. These objects should
            subclass the torch.Dataset (https://pytorch.org/docs/stable/data.html#torch.utils.data.Dataset), as this is
            what is used internally in the Trainer class. A Trainer object will be instantiated, using these dataset
            classes, inside the train method. This option exists if users want dataset functionality that is different
            than the provided default dataset classes.

        - Trainer object:
            Provide an instantiated transformers.Trainer object as input to the train class. Providing the most
            customization possible, this option exists for users who want special functionality while training (ie, by
            subclassing the transformers.Trainer class and overriding methods).

        The first two options allow the user to pass keyword arguments for training hyper-parameters (see the API
        docs for specifics), while the third allows users to instantiated their own Trainer object with what ever
        hyper-parameters they want. See examples in the use-cases sections of the documentation, or in the BERT demo
        notebooks.

    - predict():
        The predict methods provide a straightforward inference call for each task class, taking as input
        non-tokenized text for the associated task and directly outputting the model's predictions.

    - generate_outputs():
        This method is also a way to use the model for inference. However, it differs from predict() in that the outputs
        are not processed to parse a direct prediction (for example by calling argmax on output probabilities).
        Also, generate_outputs can return hidden state outputs and attention matrices for each layer in
        the base BERT model, in addition to possible fine tuning head layer outputs. Furthermore, information on how the
        input text was tokenized is returned as well. This method is useful when users want to analyze the
        outputs/network more than just getting a direct prediction. For more information, see the API docs.

In addition to these 5 methods, the task classes all expose their tokenizers and models as attributes, giving direct
access to the Huggingface tokenizer and model classes, if needed.
"""
from mangoes.modeling.bert_base import *
from mangoes.modeling.bert_finetuning import *
from mangoes.modeling.bert_pretraining import *
from mangoes.modeling.coref import *
from mangoes.modeling.tokenizer import *
from mangoes.modeling.training_utils import *
